from gmat.uvlmm.gwas.fixed import fixed
from gmat.uvlmm.gwas.fixed_ import fixed_
from gmat.uvlmm.gwas.fixed_repeat import fixed_repeat

from gmat.uvlmm.gwas.fixed_gbye import fixed_gbye
from gmat.uvlmm.gwas.fixed_min import fixed_min
